# ===================================
# AOSE sampler instance that can be used for probabilistic inference
# 
# * Translates Bayes network to ASP instance where MLE can be found
# * Constructs P(q|e) and P(-q|e) that are run on separate threads to k answer sets
# * Approximates 
# 
# USAGE
# 1. Initialize the class clingo_sampler() with number k and BN that is in form of utils.Network
# 2. Use clingo_sampler.query() with evidence, query dictionaries
# ===================================

# Utility that hold CPT:s for bayes network
from utils.network import network
# Utility that runs ASOE algorithm
from .clingo_folder.application import Application
import matplotlib.pyplot as plt
import numpy as np
import threading
import itertools
import clingo
import pickle
import queue
import time
import sys
import os


# Clingo Sampler class
# script directory needs to include /clingo_folder/ where we save all lp-files
class clingo_sampler:

    # ===================================
    # Init
    # Input:    file    - Location of bayes network as detailed in Network class  
    #           K       - Iterations,     
    #           m       - multiplier
    def __init__(self, file, K=100, m=10):
        self.clingo_file_path = os.path.join(os.path.dirname(__file__), "./clingo_folder/")
        # Make sure input is correct and files can be found
        assert os.path.isfile(file), "No network file found"
        
        # assert type(0) is network, "Input network is not of type Network"
        assert os.path.isfile(self.clingo_file_path + '/map.lp'), "No MAP encoding found"
        
        # Set up buffer for multithreading
        BUFFER_SIZE = 10000
        self.q = queue.Queue(BUFFER_SIZE)
        with open(file, 'rb') as f:
            self.cpt = pickle.load(f)

        # Parse the network to .lp
        self.transform(m)
        self.m = m
        self.K = K


    # ===================================
    # Perform a transformation on input bayes network
    # Input:    Bayes network
    # Output:   LP transformation
    def transform(self, m=10):
        # Setup a true probability lookup table
        self.true_probabilities = {}
        self.node_to_id = {}
        self.id_to_node = {}
        net_string = ""
        # Encode all nodes: node(V)
        for node, node_id in self.cpt.nodes.items():
            self.node_to_id[node] = node_id
            self.id_to_node[node_id] = node
            net_string += "node({}).\n".format(node_id)
        
        # Encode node states: states(V,S)
        for node, node_id in self.cpt.nodes.items():
            for _, state_id in self.cpt.net[node].states.items():
                net_string += "states({},{}).\n".format(node_id, state_id)
            
        # Encode node parents: parent(V,P)
        for node, node_id in self.cpt.nodes.items():
            for parent in self.cpt.net[node].parents:
                net_string += "parent({},{}).\n".format(node_id, self.cpt.nodes[parent])                

        # Encode node probabilities: state(V,P) :- assignment(..)
        for node, node_id in self.cpt.nodes.items():
            for i, (assignment, probability) in enumerate(self.cpt.net[node].cpt.items()):
                net_string += "state({},{},{}) :- ".format(node_id, i, int( -m * np.log(probability)))
                
                for parent_assignment in assignment.split(","):
                    parent, parent_val = parent_assignment.split(":")
                    net_string += "node_value({},{}), ".format(self.cpt.nodes[parent], parent_val)
                net_string = net_string[:-2] + ".\n"

                # Store the true probability, key is atom state 
                # This is better because we assume that some fraction of information is lost with NLL
                self.true_probabilities["({},{})".format(node_id, i)] = probability

        # Save the network-file
        with open(self.clingo_file_path + "net.lp", "w") as f:
            f.write(net_string)

    # ===================================
    # Makes a query LP-file that is attached to transformation
    def make_query_file(self, query, evidence, multiThread=False):
        query_string = ""
        
        # query as facts: query(V)
        for node, val in query.items():
            query_string += "query({}).\n".format(self.cpt.nodes[node])

        # evidence as facts: fix_node(V,S).
        for node, val in evidence.items():
            query_string += "fix_node({},{}).\n".format(self.cpt.nodes[node], val)

        q_files = []
        # If query is fixed (multiThread)
        if multiThread == True:
            tmp_string = query_string[:]
            for node, val in query.items():
                tmp_string += "fix_node({},{}).\n".format(self.cpt.nodes[node], val) 
            q_file = self.clingo_file_path + "query0.lp"
            with open(q_file, "w") as f:
                f.write(tmp_string)
            q_files.append(q_file)
            
            
            tmp_string = query_string[:]
            tmp_string += ":- "
            for node, val in query.items():
                tmp_string += "node_value({},{}),".format(self.cpt.nodes[node], val) 
            tmp_string = tmp_string[:-1] + ".\n"
            q_file = self.clingo_file_path + "query1.lp"
            with open(q_file, "w") as f:
                f.write(tmp_string)
            q_files.append(q_file)
            return q_files
        # Else we call the function twice
        else:
            return q_files
        
    # ===================================
    # Makes a query LP-file that 
    def query(self, query, evidence, multiThread=True):
        self.probability_estimate = []
        self.estimates = {0:0, 1:0}
        filenames = self.make_query_file(query, evidence, multiThread)
        self.done = False

        if multiThread:
            thread_list = []
            # Start all searches
            for i, filename in enumerate(filenames):
                new_thread = threading.Thread(target=self.run_clingo, args=(i, filename, self.K))
                thread_list.append(new_thread)

            # Append consumer 
            consumer_t = threading.Thread(target=self.run_consumer, args=( ))
            consumer_t.start()
            
            # Start all threads
            for t in thread_list:
                t.start()

            # Wait that all clingo threads finish
            for t in thread_list:
                t.join()
            
            # Tell consumer thread to stop
            self.done = True

        return self.probability_estimate

    # ===================================
    # Runs target clingo instance with given clingo logic file
    def run_clingo(self, i, query='clingo_folder/query.lp', N = 1000): 
        clingo_argumenst = [self.clingo_file_path + '/map.lp', query, self.clingo_file_path + '/net.lp', "-t 2",'--opt-mode=OptN', N, '--quiet=2']
        query_true = True if i == 0 else False
        clingo_app = Application(sys.argv[0], self.q, self.true_probabilities, query_true)
        clingo.clingo_main(clingo_app, clingo_argumenst)



    # ===================================
    # Runs consumer thread that keeps updating our belief for probability estimate
    def run_consumer(self):
        # Estimates:
        # 1: Model matches query
        # 0: Model does not match
        start = None

        while not self.done:
            if not self.q.empty():
                if start == None:
                    start = time.time()

                probability, model, run_time, costs, query_true = self.q.get()
                if query_true:
                    self.estimates[1] += probability
                else:
                    self.estimates[0] += probability

                self.probability_estimate.append((time.time() - start, self.estimates[1] / (self.estimates[1] + self.estimates[0])))
        
